/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Application to render text to speech with the eSpeak TTS engine
 * 
 * Copyright (C) 2006-2007, Francois Aucamp
 *
 * Francois Aucamp <faucamp@csir.co.za>
 * Neil Kleynhans <ntkleynhans@csir.co.za>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2.
 * See the COPYING file included in this archive
 */

 /*! \file
 *
 * \brief Say text to the user, using eSpeak
 * 
 * This text-to-speech application uses the eSpeak TTS engine.
 * It is designed to be used exactly like the standard 
 * "Festival" TTS application; parts of app_espeak is based on it.
 * app_espeak does however introduce some additional options, such
 * as specifying the TTS voice/language, playback speed, etc.
 * \ingroup applications
 */

/*** MODULEINFO
    <defaultenabled>no</defaultenabled>
 ***/
#define AST_MODULE "espeak"

#ifndef ASTERISK_VERSION
    /* Set this to the version of Asterisk you are compiling against:
        0 == Asterisk 1.0.x
        2 == Asterisk 1.2.x
        4 == Asterisk 1.4.x
     */
    #define ASTERISK_VERSION 4
#endif

#include <stdio.h>
#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/utils.h>
#include <asterisk/md5.h>
#include <string.h>
#include <stdlib.h>

#include <espeak/speak_lib.h>

#include <sndfile.h>
#include <samplerate.h>

#define ESPEAK_CONFIG "espeak.conf"

#if ASTERISK_VERSION < 4
static char *tdesc = "eSpeak Text-to-Speech Interface";
#endif

static char *app = "eSpeak";

static char *synopsis = "Say text to the user, using eSpeak";

static char *descrip = 
"  eSpeak(text[|intkeys]):  This will invoke the eSpeak TTS engine,\n"
"send a text string, get back the resulting waveform and play it to\n"
"the user, allowing any given interrupt keys to immediately terminate\n"
"and return\n";

#if ASTERISK_VERSION < 4
STANDARD_LOCAL_USER;
LOCAL_USER_DECL;
#endif

static int espeak_synth_callback(short *wav, int numsamples, espeak_EVENT *events)
{
    if (wav)
    {
        fwrite(wav, sizeof(short), numsamples, events[0].user_data);
    }
    return 0;
}

static int espeak_uri_callback(int type, const char *uri, const char *base)
{
    /* TODO: implement <audio> tag support.
       For now, simply say the text alternative. */
    return 1;
}

static int espeak_exec(struct ast_channel *chan, void *data)
{
    int res = 0; /* Application result */
#if ASTERISK_VERSION < 4
    struct localuser *u;
#else
    struct ast_module_user *u;
#endif
    char *intstr = "";  /* String containing DTMF string to interrupt on, can be AST_DIGIT_ANY */
    const char *tmp; /* Contains duplicate of <data>, for modification */
    FILE *fPtr; /* File ptr for eSpeak audio data */
    struct ast_config *cfg; /* Asterisk config file struct */
    int useCache;
    const char *cacheDir;
    char cacheFile[2048] = "";
    char cmd[255];
    /* For cache */
    struct MD5Context md5ctx;
    unsigned char MD5Res[16];
    char MD5Hex[33] = "";
    int fdesc = -1;
    int writeCache = 0;
    char koko[4] = "";
    int i;
    char filename[27];
    char wavFilename[27];
    int speed = 120; /* Speed of speech in words per minute; default is 170wpm */
    const char *voice;

    /* For wave file creation */
    static unsigned char wave_hdr[44] = {
        'R','I','F','F',0,0,0,0,'W','A','V','E','f','m','t',' ',
        0x10,0,0,0,1,0,1,0,  9,0x3d,0,0,0x12,0x7a,0,0,
        2,0,0x10,0,'d','a','t','a',  0,0,0,0 };
    int *p;
    int rate;

/**** NTK ****/
    float *data_wav_in, *data_wav_out;
    sf_count_t out_frames, return_size, file_size;
    SNDFILE *handle_wav_in, *handle_wav_out;
    SF_INFO info_wav_in, info_wav_out;
    SRC_DATA rate_change;
    int sr_error, status;
    char origFilename[32];
/**** NTK ****/

    if (ast_strlen_zero(data))
    {
        ast_log(LOG_WARNING, "eSpeak requires an argument (text)\n");
        return -1;
    }

#if ASTERISK_VERSION < 4
    LOCAL_USER_ADD(u);
#else
    u = ast_module_user_add(chan);
#endif

    /* Config file stuff */
#if ASTERISK_VERSION >= 2
    cfg = ast_config_load(ESPEAK_CONFIG);
#else
    cfg = ast_load(ESPEAK_CONFIG);
#endif

    if (!cfg) 
    {
        ast_log(LOG_WARNING, "eSpeak: No such configuration file %s\n", ESPEAK_CONFIG);
        useCache = 0;
        cacheDir = NULL;
        voice = "default";
    }
    else
    {
        if (!(tmp = ast_variable_retrieve(cfg, "general", "usecache")))
        {
            useCache = 0;
        }
        else
        {
            useCache = 1;
        }

        if (!(cacheDir = ast_variable_retrieve(cfg, "general", "cachedir")))
        {
            cacheDir = "/tmp";
        }

        if ((tmp = ast_variable_retrieve(cfg, "voice", "speed")))
        {
            speed = atoi(tmp);
        }
        if (!(voice = ast_variable_retrieve(cfg, "voice", "voice")))
        {
            voice = "default";
        }
    }
    /* End of config file stuff */
    
    /* Create a duplicate of the input data */
    tmp = ast_strdupa(data);

    /* Get the allowable dtmf bargein keys */
    if ((intstr = strchr(data, '|'))) 
    {
        *intstr = '\0';
        intstr++;
        if (!strcasecmp(intstr, "any"))
            intstr = AST_DIGIT_ANY;
    }

    ast_log(LOG_DEBUG, "Text passed to eSpeak: %s\n", (char *) data);

    /* Prepare the cache mechansim */
    if (useCache)
    {
        /* Compute MD5 sum of string */
        MD5Init(&md5ctx);
        MD5Update(&md5ctx,(unsigned char const *)data,strlen(data));
        MD5Final(MD5Res,&md5ctx);
        MD5Hex[0] = '\0';

        /* Convert to HEX and look if there is any matching file in the cache 
            directory */
        for (i =0; i<16; i++)
        {
            snprintf(koko, sizeof(koko), "%X", MD5Res[i]);
            strncat(MD5Hex, koko, sizeof(MD5Hex) - strlen(MD5Hex) - 1);
        }

        if (strlen(cacheDir)+strlen(MD5Hex)+5 <= 2048)
        {
            ast_log(LOG_DEBUG, "eSpeak: Activating cache mechanism...");
            snprintf(cacheFile, sizeof(cacheFile), "%s/%s.wav", cacheDir, MD5Hex);
            fdesc = open(cacheFile, O_RDONLY);
            cacheFile[(strlen(cacheFile)-4)] = '\0';
            if (fdesc == -1)
            {
                ast_log(LOG_DEBUG, "eSpeak: Cache file does not yet exist.");
                writeCache = 1;
            }
            else
            {
                close(fdesc);
                ast_log(LOG_DEBUG,"eSpeak: Cache file exists.");
                res = ast_streamfile(chan, cacheFile, chan->language);
        
                if (!res)
                {
                    res = ast_waitstream(chan, intstr);
                    ast_stopstream(chan);
                }
                else
                {
                    ast_log(LOG_WARNING, "ast_streamfile failed on %s for eSpeak\n", chan->name);
                }

#if ASTERISK_VERSION < 4
                LOCAL_USER_REMOVE(u);
#else
                ast_module_user_remove(u);
#endif
                return res;
            }
        }
    }
    /* End of cache mechanism */

    /* Create temp filename */
    sprintf(filename, "/tmp/espeak_buf_XXXXXX");
    mkstemp(filename);
    strcpy(wavFilename, filename);
    strcat(wavFilename, ".wav");
    sprintf(cmd, "%s_orig.wav", filename); /*espeak_buf_XXXXXX_orig.wav*/

    /* Invoke eSpeak */
    rate = espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS, 100, NULL, 0);
    espeak_SetSynthCallback(espeak_synth_callback);
    espeak_SetUriCallback(espeak_uri_callback);
    /* Set options */
    espeak_SetVoiceByName(voice);
    espeak_SetParameter(espeakRATE, speed, 0);
    
    /* Prepare wave file */
    fPtr = fopen(cmd, "wb");
    if(fPtr != NULL)
    {
        /* Set the sample rate in the header */
        p = (int *)(&wave_hdr[24]);
        p[0] = rate;
        p[1] = rate * 2;
        fwrite(wave_hdr, 1, sizeof(wave_hdr), fPtr);
    }
    else
    {
        ast_log(LOG_ERROR, "eSpeak: Failed to create audio buffer file '%s'\n", filename);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        return res;
    }

    /* Do the actual synthesis */
    espeak_Synth((const void *)data, (sizeof(char)*strlen((char *)data)), 0, POS_CHARACTER, 0, espeakCHARS_AUTO, NULL, fPtr);

/**** NTK ****/

    /*Fix up wav header*/
    file_size = ftell(fPtr);

    /*enter data in*/
    p = (int *)(&wave_hdr[4]);
    p[0] = (int)file_size-8;
    p = (int *)(&wave_hdr[40]);
    p[0] = (int)file_size-44;

    fseek(fPtr, 0, SEEK_SET);
    fwrite(wave_hdr, 1, sizeof(wave_hdr), fPtr);
    fclose(fPtr);

    sprintf(origFilename, "%s", cmd);

    /*open sound file*/
    handle_wav_in = sf_open(cmd, SFM_READ, &info_wav_in);

    if(!handle_wav_in)
    {
        ast_log(LOG_ERROR, "libsndfile: Failed to open %s\n", cmd);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        return res;
    }

    info_wav_in.frames = (file_size-44)/2;
    /*check channel number - will only do 1 ... for now*/
/*    if(info_wav_in.channels > 1)
    {
        ast_log(LOG_ERROR, "libsndfile: Sound file %s has more than 1 channel\n", cmd);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        return res;
    }
*/
    data_wav_in = NULL;
    data_wav_in = (float *)malloc(info_wav_in.frames * sizeof(float));

    if(!data_wav_in)
    {
        ast_log(LOG_ERROR, "eSpeak: Unable to allocate memory for %s input data\n", cmd);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        return res;
    }

    out_frames = info_wav_in.frames * (sf_count_t)8000;
    out_frames /= (sf_count_t)rate; /*info_wav_in.samplerate;*/

    data_wav_out = NULL;
    data_wav_out = (float *)malloc(out_frames * sizeof(float));

    if(!data_wav_out)
    {
        ast_log(LOG_ERROR, "eSpeak: Unable to allocate memory for %s output data\n", wavFilename);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        free(data_wav_in);
        return res;
    }

    return_size = sf_readf_float(handle_wav_in, data_wav_in, info_wav_in.frames);

    if(return_size != info_wav_in.frames)
    {
        ast_log(LOG_ERROR, "libsndfile: Requested to read %ld frames but read %ld bytes\n", info_wav_in.frames, return_size);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        free(data_wav_in);
        free(data_wav_out);
        return res;
    }

    rate_change.data_in = data_wav_in;
    rate_change.data_out = data_wav_out;
    rate_change.input_frames = info_wav_in.frames;
    rate_change.output_frames = out_frames;
    rate_change.src_ratio = (double) ((double)8000/(double)rate);
    /*
    SRC_SINC_BEST_QUALITY       = 0,
    SRC_SINC_MEDIUM_QUALITY     = 1,
    SRC_SINC_FASTEST            = 2,
    SRC_ZERO_ORDER_HOLD         = 3,
    SRC_LINEAR                  = 4
    */

    sr_error = src_simple(&rate_change, SRC_LINEAR, 1) ;

    if(sr_error != 0)
    {
        ast_log(LOG_ERROR, "libsamplerate: %s", src_strerror(sr_error));
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        free(data_wav_in);
        free(data_wav_out);
        return res;
    }

    memcpy((void *)&info_wav_out,(const void *) &info_wav_in, sizeof(SF_INFO));
    info_wav_out.samplerate = 8000;
    info_wav_out.frames = out_frames;
    handle_wav_out = sf_open(wavFilename, SFM_WRITE, &info_wav_out);

    if(!handle_wav_out)
    {
        ast_log(LOG_ERROR, "libsndfile: Failed to open %s\n", wavFilename);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        free(data_wav_in);
        free(data_wav_out);
        return res;
    }

    return_size = sf_writef_float(handle_wav_out, data_wav_out, out_frames);

    if(return_size != out_frames)
    {
        ast_log(LOG_ERROR, "libsndfile: Requested to write out %ld frames but wrote %ld frames\n", out_frames, return_size);
#if ASTERISK_VERSION < 4
        LOCAL_USER_REMOVE(u);
#else
        ast_module_user_remove(u);
#endif
        sf_close(handle_wav_in);
        sf_close(handle_wav_out);
        free(data_wav_in);
        free(data_wav_out);
        return res;
    }

    sf_close(handle_wav_in);
    sf_close(handle_wav_out);
    free(data_wav_in);
    free(data_wav_out);

/**** NTK ****/

    /* Convert wave data to 8000Hz using sox */ 
    /*sprintf(cmd, "sox %s_orig.wav -r 8000 %s", filename, wavFilename);
    system(cmd);*/

    if (chan->_state != AST_STATE_UP)
    {
        /* Answer the channel if it is down */
        ast_answer(chan);
    }

    /* Save audio file to cache, if required */
    if (writeCache)
    {
        ast_log(LOG_DEBUG, "eSpeak: Creating cache file...");
        sprintf(cmd, "cp -f %s %s.wav", wavFilename, cacheFile);
        system(cmd);
    }

    /* Play back the audio file */
    ast_stopstream(chan);
    res = ast_streamfile(chan, filename, chan->language);
    if (!res)
    {
        res = ast_waitstream(chan, intstr);
        ast_stopstream(chan);
    }
    else
    {
        ast_log(LOG_WARNING, "ast_streamfile failed on %s for eSpeak\n", chan->name);
    }

    /* Remove temp file */
    /*sprintf(cmd, "rm -f %s*", filename);
    system(cmd);*/
    status = remove(filename);
    status = remove(wavFilename);
    status = remove(origFilename);

#if ASTERISK_VERSION < 4
    LOCAL_USER_REMOVE(u);
#else
    ast_module_user_remove(u);
#endif
    return res;
}

#if ASTERISK_VERSION == 4
static int unload_module(void)
{
    int res;
    res = ast_unregister_application(app);
    return res; 
}

static int load_module(void)
{
    return ast_register_application(app, espeak_exec, synopsis, descrip);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "eSpeak TTS Interface");

#else
int unload_module(void)
{
    int res;
    res = ast_unregister_application(app);
    STANDARD_HANGUP_LOCALUSERS;
    return res; 
}

int load_module(void)
{
    return ast_register_application(app, espeak_exec, synopsis, descrip);
}


char *description(void)
{
    return tdesc;
}

int usecount(void)
{
    int res;
    STANDARD_USECOUNT(res);
    return res;
}

char *key()
{
    return ASTERISK_GPL_KEY;
}
#endif
